{% extends "!layout.html" %} {% block menu %} {{ super() }}
<a href="{{pathto('py-modindex.html', 1)}}">Module Index (API)</a>
<a href="{{pathto('genindex.html', 1)}}">General Index</a>
{% endblock %}
